@@ -90,6 +90,7 @@
 (defvar-local exwm--fullscreen nil)       ;used in fullscreen
 (defvar-local exwm--floating-frame-position nil) ;used in fullscreen
 (defvar-local exwm--fixed-size nil)              ;fixed size
+(defvar-local exwm--keyboard-grabbed nil)        ;Keyboard grabbed.
 (defvar-local exwm--on-KeyPress         ;KeyPress event handler
   #'exwm-input--on-KeyPress-line-mode)
 ;; Properties
@@ -101,7 +102,7 @@
 (defvar-local exwm--title-is-utf8 nil)
 (defvar-local exwm-transient-for nil "WM_TRANSIENT_FOR.")
 (defvar-local exwm--protocols nil)
-(defvar-local exwm-state nil "WM_STATE.")
+(defvar-local exwm-state xcb:icccm:WM_STATE:NormalState "WM_STATE.")
 ;; _NET_WM_NORMAL_HINTS
 (defvar-local exwm--normal-hints-x nil)
 (defvar-local exwm--normal-hints-y nil)
@@ -118,6 +119,15 @@
 ;; _MOTIF_WM_HINTS
 (defvar-local exwm--mwm-hints-decorations t)
 
+(declare-function exwm-floating-hide "exwm-floating.el")
+(declare-function exwm-floating-toggle-floating "exwm-floating.el")
+(declare-function exwm-input-release-keyboard "exwm-input.el")
+(declare-function exwm-input-send-next-key "exwm-input.el" (times))
+(declare-function exwm-layout-set-fullscreen "exwm-layout.el" (&optional id))
+(declare-function exwm-layout-toggle-mode-line "exwm-layout.el")
+(declare-function exwm-workspace-move-window "exwm-workspace.el"
+                  (frame-or-index &optional id))
+
 (defvar exwm-mode-map
   (let ((map (make-sparse-keymap)))
     (define-key map "\C-c\C-f" #'exwm-layout-set-fullscreen)
@@ -130,6 +140,78 @@
     map)
   "Keymap for `exwm-mode'.")
 
+;; This menu mainly acts as an reminder for users.  Thus it should be as
+;; detailed as possible, even some entries do not make much sense here.
+;; Also, inactive entries should be disabled rather than hidden.
+(easy-menu-define exwm-mode-menu exwm-mode-map
+  "Menu for `exwm-mode'."
+  '("EXWM"
+    "---"
+    "*General*"
+    "---"
+    ["Toggle floating" exwm-floating-toggle-floating]
+    ["Enter fullscreen" exwm-layout-set-fullscreen (not exwm--fullscreen)]
+    ["Leave fullscreen" exwm-reset exwm--fullscreen]
+    ["Hide window" exwm-floating-hide exwm--floating-frame]
+
+    "---"
+    "*Resizing*"
+    "---"
+    ["Toggle mode-line" exwm-layout-toggle-mode-line :keys "C-c C-t C-m"]
+    ["Enlarge window vertically" exwm-layout-enlarge-window]
+    ["Enlarge window horizontally" exwm-layout-enlarge-window-horizontally]
+    ["Shrink window vertically" exwm-layout-shrink-window]
+    ["Shrink window horizontally" exwm-layout-shrink-window-horizontally]
+
+    "---"
+    "*Keyboard*"
+    "---"
+    ["Capture keyboard" exwm-input-release-keyboard exwm--keyboard-grabbed]
+    ;; It's recommended to use `exwm-reset' rather than
+    ;; `exwm-input-grab-keyboard' to release keyboard (enter line-mode).
+    ["Release keyboard" exwm-reset (not exwm--keyboard-grabbed)]
+    ["Send key" exwm-input-send-next-key exwm--keyboard-grabbed]
+    ;; This is merely a reference.
+    ("Send simulation key" :filter
+     (lambda (&rest _args)
+       (mapcar (lambda (i)
+                 (let ((keys (cdr i)))
+                   (if (vectorp keys)
+                       (setq keys (append keys))
+                     (unless (sequencep keys)
+                       (setq keys (list keys))))
+                   (vector (key-description keys)
+                           `(lambda ()
+                              (interactive)
+                              (dolist (key ',keys)
+                                (exwm-input--fake-key key)))
+                           :keys (key-description (car i)))))
+               exwm-input--simulation-keys)))
+
+    ["Define global binding" exwm-input-set-key]
+
+    "---"
+    "*Workspace*"
+    "---"
+    ["Add workspace" exwm-workspace-add]
+    ["Delete current workspace" exwm-workspace-delete]
+    ["Move workspace to" exwm-workspace-move]
+    ["Swap workspaces" exwm-workspace-swap]
+    ["Move X window to" exwm-workspace-move-window :keys "C-c C-m"]
+    ["Move X window from" exwm-workspace-switch-to-buffer]
+    ["Toggle minibuffer" exwm-workspace-toggle-minibuffer]
+    ["Switch workspace" exwm-workspace-switch]
+    ;; Place this entry at bottom to avoid selecting others by accident.
+    ("Switch to" :filter
+     (lambda (&rest _args)
+       (mapcar (lambda (i)
+                 `[,(format "workspace %d" i)
+                   (lambda ()
+                     (interactive)
+                     (exwm-workspace-switch ,i))
+                   (/= ,i exwm-workspace-current-index)])
+               (number-sequence 0 (1- (exwm-workspace--count))))))))
+
 (declare-function exwm-manage--kill-buffer-query-function "exwm-manage.el")
 
 (define-derived-mode exwm-mode nil "EXWM"
@@ -140,8 +222,7 @@
   (setq mode-name
         '(:eval (propertize "EXWM" 'face
                             (when (cl-some (lambda (i)
-                                             (frame-parameter i
-                                                              'exwm--urgency))
+                                             (frame-parameter i 'exwm-urgency))
                                            exwm-workspace--list)
                               'font-lock-warning-face))))
   ;; Change major-mode is not allowed
