@@ -76,72 +76,110 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
                            :revert-to xcb:InputFocus:PointerRoot
                            :focus id
                            :time xcb:Time:CurrentTime)))
+      (exwm-input--set-active-window id)
       (xcb:flush exwm--connection))))
 
-(defvar exwm-input--focus-window nil "The (Emacs) window to be focused.")
-(defvar exwm-input--timer nil "Currently running timer.")
+(defvar exwm-input--update-focus-window nil "The (Emacs) window to be focused.
+
+This value should always be overwritten.")
 
 (defun exwm-input--on-buffer-list-update ()
-  "Run in buffer-list-update-hook to track input focus."
-  (let ((frame (selected-frame))
-        (window (selected-window))
-        (buffer (current-buffer)))
-    (when (and (not (minibufferp buffer))
-               (frame-parameter frame 'exwm-outer-id) ;e.g. emacsclient frame
-               (eq buffer (window-buffer))) ;e.g. `with-temp-buffer'
-      (when exwm-input--timer (cancel-timer exwm-input--timer))
-      (setq exwm-input--focus-window window
-            exwm-input--timer
-            (run-with-idle-timer 0.01 nil #'exwm-input--update-focus)))))
+  "Run in `buffer-list-update-hook' to track input focus."
+  (when (and (not (minibufferp)) ;Do not set input focus on minibuffer window.
+             (eq (current-buffer) (window-buffer)) ;e.g. `with-temp-buffer'.
+             (frame-parameter nil 'exwm-outer-id)) ;e.g. emacsclient frame.
+    (setq exwm-input--update-focus-window (selected-window))
+    (exwm-input--update-focus-defer)))
+
+;; Input focus update requests should be accumulated for a short time
+;; interval so that only the last one need to be processed.  This not
+;; improves the overall performance, but avoids the problem of input
+;; focus loop, which is a result of the interaction with Emacs frames.
+;;
+;; FIXME: The time interval is hard to decide and perhaps machine-dependent.
+;;        A value too small can cause redundant updates of input focus,
+;;        and even worse, dead loops.  OTOH a large value would bring
+;;        laggy experience.
+(defconst exwm-input--update-focus-interval 0.01
+  "Time interval (in seconds) for accumulating input focus update requests.")
+
+(defvar exwm-input--update-focus-lock nil
+  "Lock for solving input focus update contention.")
+(defvar exwm-input--update-focus-defer-timer nil "Timer for polling the lock.")
+(defvar exwm-input--update-focus-timer nil
+  "Timer for deferring the update of input focus.")
+
+(defun exwm-input--update-focus-defer ()
+  "Defer updating input focus."
+  (when exwm-input--update-focus-defer-timer
+    (cancel-timer exwm-input--update-focus-defer-timer))
+  (if exwm-input--update-focus-lock
+      (setq exwm-input--update-focus-defer-timer
+            (run-with-idle-timer 0 nil
+                                 #'exwm-input--update-focus-defer))
+    (setq exwm-input--update-focus-defer-timer nil)
+    (when exwm-input--update-focus-timer
+      (cancel-timer exwm-input--update-focus-timer))
+    (setq exwm-input--update-focus-timer
+          (run-with-idle-timer exwm-input--update-focus-interval nil
+                               #'exwm-input--update-focus
+                               exwm-input--update-focus-window))))
 
 (defvar exwm-workspace--current)
 (defvar exwm-workspace--switch-history-outdated)
 (defvar exwm-workspace-current-index)
 (defvar exwm-workspace--minibuffer)
 
-(defun exwm-input--update-focus ()
+(declare-function exwm-layout--iconic-state-p "exwm-layout.el" (&optional id))
+(declare-function exwm-layout--set-state "exwm-layout.el" (id state))
+(declare-function exwm-workspace--minibuffer-own-frame-p "exwm-workspace.el")
+(declare-function exwm-workspace-switch "exwm-workspace.el"
+                  (frame-or-index &optional force))
+
+(defun exwm-input--update-focus (window)
   "Update input focus."
-  (when (and (window-live-p exwm-input--focus-window)
+  (setq exwm-input--update-focus-lock t)
+  (when (and (window-live-p window)
              ;; Do not update input focus when there's an active minibuffer.
              (not (active-minibuffer-window)))
-    (with-current-buffer (window-buffer exwm-input--focus-window)
+    (with-current-buffer (window-buffer window)
       (if (eq major-mode 'exwm-mode)
           (if (not (eq exwm--frame exwm-workspace--current))
-              ;; Do not focus X windows on other workspace
+              ;; Do not focus X windows on other workspace.
               (progn
-                (set-frame-parameter exwm--frame 'exwm--urgency t)
+                (set-frame-parameter exwm--frame 'exwm-urgency t)
                 (setq exwm-workspace--switch-history-outdated t)
                 (force-mode-line-update)
                 ;; The application may have changed its input focus
-                (exwm-workspace-switch exwm-workspace-current-index t))
+                (select-window
+                 (frame-selected-window exwm-workspace--current)))
             (exwm--log "Set focus on #x%x" exwm--id)
             (exwm-input--set-focus exwm--id)
             (when exwm--floating-frame
               ;; Adjust stacking orders of the floating container.
-              (if (exwm-workspace--minibuffer-own-frame-p)
-                  ;; Put this floating X window just below the minibuffer.
-                  (xcb:+request exwm--connection
-                      (make-instance 'xcb:ConfigureWindow
-                                     :window exwm--container
-                                     :value-mask
-                                     (logior xcb:ConfigWindow:Sibling
-                                             xcb:ConfigWindow:StackMode)
-                                     :sibling (frame-parameter
-                                               exwm-workspace--minibuffer
-                                               'exwm-container)
-                                     :stack-mode xcb:StackMode:Below))
-                ;; Put this floating X window at top.
-                (xcb:+request exwm--connection
-                    (make-instance 'xcb:ConfigureWindow
-                                   :window exwm--container
-                                   :value-mask xcb:ConfigWindow:StackMode
-                                   :stack-mode xcb:StackMode:Above)))
+              (xcb:+request exwm--connection
+                  (make-instance 'xcb:ConfigureWindow
+                                 :window exwm--container
+                                 :value-mask xcb:ConfigWindow:StackMode
+                                 :stack-mode xcb:StackMode:Above))
+              ;; This floating X window might be hide by `exwm-floating-hide'.
+              (when (exwm-layout--iconic-state-p)
+                (exwm-layout--set-state exwm--id
+                                        xcb:icccm:WM_STATE:NormalState))
               (xcb:flush exwm--connection)))
-        (when (eq (selected-window) exwm-input--focus-window)
-          (exwm--log "Focus on %s" exwm-input--focus-window)
-          (select-frame-set-input-focus (window-frame exwm-input--focus-window)
-                                        t)))
-      (setq exwm-input--focus-window nil))))
+        (when (eq (selected-window) window)
+          (exwm--log "Focus on %s" window)
+          (select-frame-set-input-focus (window-frame window) t)
+          (exwm-input--set-active-window)
+          (xcb:flush exwm--connection)))))
+  (setq exwm-input--update-focus-lock nil))
+
+(defun exwm-input--set-active-window (&optional id)
+  "Set _NET_ACTIVE_WINDOW."
+  (xcb:+request exwm--connection
+      (make-instance 'xcb:ewmh:set-_NET_ACTIVE_WINDOW
+                     :window exwm--root
+                     :data (or id xcb:Window:None))))
 
 (defvar exwm-input--during-key-sequence nil
   "Non-nil indicates Emacs is waiting for more keys to form a key sequence.")
@@ -159,49 +197,57 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
 
 (declare-function exwm-floating--start-moveresize "exwm-floating.el"
                   (id &optional type))
+(declare-function exwm-workspace--position "exwm-workspace.el" (frame))
+(declare-function exwm-workspace--workspace-p "exwm-workspace.el" (workspace))
 
 (defvar exwm-workspace--list)
 
 (defun exwm-input--on-ButtonPress (data _synthetic)
   "Handle ButtonPress event."
   (let ((obj (make-instance 'xcb:ButtonPress))
-        (mode xcb:Allow:SyncPointer))
+        (mode xcb:Allow:SyncPointer)
+        window buffer frame)
     (xcb:unmarshal obj data)
     (with-slots (detail time event state) obj
       (setq exwm-input--timestamp time)
+      (setq window (get-buffer-window (exwm--id->buffer event) t)
+            buffer (window-buffer window))
       (cond ((and (= state exwm-input--move-mask)
-                  (= detail exwm-input--move-keysym))
+                  (= detail exwm-input--move-keysym)
+                  ;; Either an undecorated or a floating X window.
+                  (with-current-buffer buffer
+                    (or (not (eq major-mode 'exwm-mode))
+                        exwm--floating-frame)))
              ;; Move
              (exwm-floating--start-moveresize
               event xcb:ewmh:_NET_WM_MOVERESIZE_MOVE))
             ((and (= state exwm-input--resize-mask)
-                  (= detail exwm-input--resize-keysym))
+                  (= detail exwm-input--resize-keysym)
+                  (with-current-buffer buffer
+                    (or (not (eq major-mode 'exwm-mode))
+                        exwm--floating-frame)))
              ;; Resize
              (exwm-floating--start-moveresize event))
             (t
              ;; Click to focus
-             (let ((window (get-buffer-window (exwm--id->buffer event) t))
-                   frame)
-               (unless (eq window (selected-window))
-                 (setq frame (window-frame window))
-                 (unless (eq frame exwm-workspace--current)
-                   (if (memq frame exwm-workspace--list)
-                       ;; The X window is on another workspace
-                       (exwm-workspace-switch
-                        (cl-position frame exwm-workspace--list))
-                     (with-current-buffer (window-buffer window)
-                       (when (and (eq major-mode 'exwm-mode)
-                                  (not (eq exwm--frame
-                                           exwm-workspace--current)))
-                         ;; The floating X window is on another workspace
-                         (exwm-workspace-switch
-                          (cl-position exwm--frame exwm-workspace--list))))))
-                 ;; It has been reported that the `window' may have be deleted
-                 (if (window-live-p window)
-                     (select-window window)
-                   (setq window
-                         (get-buffer-window (exwm--id->buffer event) t))
-                   (when window (select-window window)))))
+             (unless (eq window (selected-window))
+               (setq frame (window-frame window))
+               (unless (eq frame exwm-workspace--current)
+                 (if (exwm-workspace--workspace-p frame)
+                     ;; The X window is on another workspace
+                     (exwm-workspace-switch frame)
+                   (with-current-buffer buffer
+                     (when (and (eq major-mode 'exwm-mode)
+                                (not (eq exwm--frame
+                                         exwm-workspace--current)))
+                       ;; The floating X window is on another workspace
+                       (exwm-workspace-switch exwm--frame)))))
+               ;; It has been reported that the `window' may have be deleted
+               (if (window-live-p window)
+                   (select-window window)
+                 (setq window
+                       (get-buffer-window (exwm--id->buffer event) t))
+                 (when window (select-window window))))
              ;; The event should be replayed
              (setq mode xcb:Allow:ReplayPointer))))
     (xcb:+request exwm--connection
@@ -268,7 +314,22 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
   "Set a global key binding."
   (interactive "KSet key globally: \nCSet key %s to command: ")
   (global-set-key key command)
-  (cl-pushnew key exwm-input--global-keys))
+  (cl-pushnew key exwm-input--global-keys)
+  (when (called-interactively-p 'any)
+    (exwm-input--update-global-prefix-keys)))
+
+;; FIXME: Putting (t . EVENT) into `unread-command-events' does not really work
+;;        as documented in Emacs 24.  Since inserting a conventional EVENT does
+;;        add it into (this-command-keys) there, we use `unread-command-events'
+;;        differently on Emacs 24 and 25.
+(eval-and-compile
+  (if (< emacs-major-version 26)
+      (defsubst exwm-input--unread-event (event)
+        (setq unread-command-events
+              (append unread-command-events (list event))))
+    (defsubst exwm-input--unread-event (event)
+      (setq unread-command-events
+            (append unread-command-events `((t . ,event)))))))
 
 (defvar exwm-input-command-whitelist nil
   "A list of commands that when active all keys should be forwarded to Emacs.")
@@ -294,7 +355,9 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
                      (memq event exwm-input--simulation-prefix-keys)))
         (setq mode xcb:Allow:AsyncKeyboard)
         (unless minibuffer-window (setq exwm-input--during-key-sequence t))
-        (push event unread-command-events))
+        ;; Feed this event to command loop.  Also force it to be added to
+        ;; `this-command-keys'.
+        (exwm-input--unread-event event))
       (xcb:+request exwm--connection
           (make-instance 'xcb:AllowEvents
                          :mode (or mode xcb:Allow:ReplayKeyboard)
@@ -317,7 +380,8 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
           (setq exwm-input--temp-line-mode t
                 exwm-input--during-key-sequence t)
           (exwm-input--grab-keyboard))  ;grab keyboard temporarily
-        (push event unread-command-events))))
+        (setq unread-command-events
+              (append unread-command-events (list event))))))
   (xcb:+request exwm--connection
       (make-instance 'xcb:AllowEvents
                      :mode xcb:Allow:AsyncKeyboard
@@ -388,6 +452,7 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
   (when id
     (with-current-buffer (exwm--id->buffer id)
       (exwm-input--grab-keyboard id)
+      (setq exwm--keyboard-grabbed t)
       (exwm-input--update-mode-line id)
       (force-mode-line-update))))
 
@@ -398,6 +463,7 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
   (when id
     (with-current-buffer (exwm--id->buffer id)
       (exwm-input--release-keyboard id)
+      (setq exwm--keyboard-grabbed nil)
       (exwm-input--update-mode-line id)
       (force-mode-line-update))))
 
@@ -443,6 +509,8 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
           (setq key (read-key (format "Send key: %s (%d/%d)"
                                       (key-description keys)
                                       (1+ i) times)))
+          (when (and (listp key) (eq (car key) t))
+            (setq key (cdr key)))
           (unless (listp key) (throw 'break nil))))
       (setq exwm-input--during-key-sequence nil)
       (setq keys (vconcat keys (vector key)))
@@ -453,11 +521,16 @@ It's updated in several occasions, and only used by `exwm-input--set-focus'.")
 ;;   (unless (listp last-input-event)      ;not a key event
 ;;     (exwm-input--fake-key last-input-event)))
 
+(defvar exwm-input--local-simulation-keys nil
+  "Whether simulation keys are local.")
+
 (defun exwm-input--update-simulation-prefix-keys ()
   "Update the list of prefix keys of simulation keys."
   (setq exwm-input--simulation-prefix-keys nil)
   (dolist (i exwm-input--simulation-keys)
-    (define-key exwm-mode-map (car i) #'exwm-input-send-simulation-key)
+    (if exwm-input--local-simulation-keys
+        (local-set-key (car i) #'exwm-input-send-simulation-key)
+      (define-key exwm-mode-map (car i) #'exwm-input-send-simulation-key))
     (cl-pushnew (elt (car i) 0) exwm-input--simulation-prefix-keys)))
 
 (defun exwm-input-set-simulation-keys (simulation-keys)
@@ -469,6 +542,16 @@ SIMULATION-KEYS is an alist of the form (original-key . simulated-key)."
     (cl-pushnew `(,(vconcat (car i)) . ,(cdr i)) exwm-input--simulation-keys))
   (exwm-input--update-simulation-prefix-keys))
 
+(defun exwm-input-set-local-simulation-keys (simulation-keys)
+  "Set buffer-local simulation keys.
+
+Its usage is the same with `exwm-input-set-simulation-keys'."
+  (make-local-variable 'exwm-input--simulation-keys)
+  (make-local-variable 'exwm-input--simulation-prefix-keys)
+  (use-local-map (copy-keymap exwm-mode-map))
+  (let ((exwm-input--local-simulation-keys t))
+    (exwm-input-set-simulation-keys simulation-keys)))
+
 ;;;###autoload
 (defun exwm-input-send-simulation-key (times)
   "Fake a key event according to last input key sequence."
