@@ -65,6 +65,11 @@ You shall use the default value if using auto-hide minibuffer.")
   "The selection owner window.")
 (defvar exwm-systemtray--embedder nil "The embedder window.")
 
+(defvar exwm-workspace--current)
+(declare-function exwm-workspace--current-height "exwm-workspace.el")
+(declare-function exwm-workspace--current-width  "exwm-workspace.el")
+(declare-function exwm-workspace--minibuffer-own-frame-p "exwm-workspace.el")
+
 (defun exwm-systemtray--embed (icon)
   "Embed an icon."
   (exwm--log "(System Tray) Try to embed #x%x" icon)
@@ -266,8 +271,6 @@ You shall use the default value if using auto-hide minibuffer.")
               (t
                (exwm--log "(System Tray) Unknown opcode message: %s" obj)))))))
 
-(defvar exwm-workspace--current)
-
 (defun exwm-systemtray--on-workspace-switch ()
   "Reparent/Refresh the system tray in `exwm-workspace-switch-hook'."
   (unless (exwm-workspace--minibuffer-own-frame-p)
@@ -296,7 +299,7 @@ You shall use the default value if using auto-hide minibuffer.")
 (defvar xcb:Atom:_NET_SYSTEM_TRAY_S0)
 (defvar exwm-workspace--minibuffer)
 
-(defun exwm-systemtray--init ()
+(cl-defun exwm-systemtray--init ()
   "Initialize system tray module."
   (cl-assert (not exwm-systemtray--connection))
   (cl-assert (not exwm-systemtray--list))
@@ -319,7 +322,9 @@ You shall use the default value if using auto-hide minibuffer.")
           (make-instance 'xcb:GetSelectionOwner
                          :selection xcb:Atom:_NET_SYSTEM_TRAY_S0))
     (when (/= owner xcb:Window:None)
-      (error "[EXWM] Other system tray detected")))
+      (xcb:disconnect exwm-systemtray--connection)
+      (warn "[EXWM] Other system tray detected")
+      (cl-return-from exwm-systemtray--init)))
   (let ((id (xcb:generate-id exwm-systemtray--connection)))
     (setq exwm-systemtray--selection-owner-window id)
     (xcb:+request exwm-systemtray--connection
